'''
That’s a comprehensive set of questions! Here’s a detailed explanation for each:  

---

### **a) K-Nearest Neighbors (KNN) Algorithm**  
KNN is a non-parametric, instance-based learning algorithm used for classification and regression tasks. It classifies a new data point based on the majority class among its **K** nearest neighbors in feature space.  

#### **Example:**  
Suppose we want to classify whether a fruit is an apple or an orange based on its weight and texture. We store previous fruit data and use KNN to determine the label of a new fruit by checking the majority class among its nearest **K** neighbors.  

### **b)**
#### **i) F1 Score**  
The **F1 Score** balances **Precision** and **Recall** and is calculated as:  
\[
F1 = 2 \times \frac{Precision \times Recall}{Precision + Recall}
\]

#### **ii) Precision Calculation from Confusion Matrix**
|                 | Predicted 0 | Predicted 1 |
|---------------|------------|------------|
| **Actual 0**  | 56850      | 4          |
| **Actual 1**  | 30         | 78         |

\[
Precision = \frac{TP}{TP + FP} = \frac{78}{78 + 4} = 0.9512 \ (\approx 95.12\%)
\]

---

### **c) AdaBoost vs. Gradient Boosting**  
- **AdaBoost (Adaptive Boosting):**  
  - Sequentially trains weak classifiers by assigning higher weights to misclassified samples.  
  - Uses decision stumps (shallow trees).  
- **Gradient Boosting:**  
  - Uses gradient descent to minimize errors.  
  - Fits new models on residual errors of previous models.  

**Differences:**  
| Feature | AdaBoost | Gradient Boosting |
|---------|---------|------------------|
| Weak Learners | Decision stumps | Deeper trees |
| Error Handling | Adjusts sample weights | Uses gradient descent |
| Performance | Faster, less prone to overfitting | Slower, more accurate |

**Advantages:**  
- AdaBoost is simple and effective for small datasets.  
- Gradient Boosting is highly flexible and improves complex models.

---

### **d) Stacking Classifier**  
A meta-learning technique where multiple models’ outputs are combined using a final model (meta-classifier).  
- **Advantages:**  
  - Improves accuracy.  
  - Reduces overfitting by using diverse base models.  

---

### **e) Decision Tree Working**  
A hierarchical model that splits data based on feature conditions.  
#### **Example:**  
Predicting if a person buys a car based on income and credit history.  
1. Root Node: Income  
2. Split: High income → Yes, Low income → Check credit history  
3. Further Split: Good credit → Yes, Poor credit → No  

---

### **f) Decision Tree vs. Random Forest**  
| **Feature** | **Decision Tree** | **Random Forest** |
|------------|----------------|----------------|
| Overfitting | High | Low |
| Speed | Fast | Slower |
| Accuracy | Moderate | High |
| Feature Importance | Uses one tree | Averages multiple trees |

---

### **g) Precision & Recall**  
- **Precision:** \(\frac{TP}{TP + FP}\) → Out of all predicted positives, how many are correct?  
- **Recall:** \(\frac{TP}{TP + FN}\) → Out of all actual positives, how many are correctly predicted?  

---

### **h) Bagging vs. Boosting**  
- **Bagging (Bootstrap Aggregation):** Trains multiple models in parallel (e.g., Random Forest).  
- **Boosting:** Trains models sequentially, focusing on misclassified samples (e.g., AdaBoost).  

---

### **i) ROC Curve**  
A graphical representation of **True Positive Rate** vs. **False Positive Rate**, used to evaluate classifiers.

---

### **j) Gini Impurity & Entropy**  
- **Gini Impurity**: Probability of incorrect classification.  
- **Entropy**: Measures randomness in a dataset.  
- **Formulas:**  
  \[
  Gini = 1 - \sum p_i^2
  \]
  \[
  Entropy = -\sum p_i \log_2 p_i
  \]

---

### **k) Random Forest**  
An ensemble of decision trees where each tree is trained on random subsets.  
**Advantages:**  
1. Reduces overfitting.  
2. Handles missing values well.  

---

### **l) Ensemble Learning**  
A technique combining multiple models (e.g., Bagging, Boosting, Stacking) to improve accuracy.

---

### **m) Decision Tree Algorithm Example**  
Predicting if a student will pass based on study hours and attendance.  
- **Root Node:** Study hours  
- **Split:** > 4 hrs → Likely to pass, < 4 hrs → Check attendance  

---

### **n) Confusion Matrix & Metrics**  
|        | Predicted Positive | Predicted Negative |
|--------|-------------------|-------------------|
| Actual Positive | TP | FN |
| Actual Negative | FP | TN |

**Metrics:**  
- **Accuracy**: \(\frac{TP + TN}{Total}\)  
- **Precision**: \(\frac{TP}{TP + FP}\)  
- **Recall**: \(\frac{TP}{TP + FN}\)  
- **F1 Score**: \(2 \times \frac{Precision \times Recall}{Precision + Recall}\)

---

### **o) Assumptions of Logistic Regression & Naïve Bayes**  
- **Logistic Regression:**  
  - No multicollinearity.  
  - Log-odds linearity.  
- **Naïve Bayes:**  
  - Assumes feature independence.  

---

### **p) Advantages & Disadvantages of ML Algorithms**  
| Algorithm | Advantages | Disadvantages |
|------------|----------------|----------------|
| Random Forest | Robust to overfitting | Slow |
| AdaBoost | Fast | Sensitive to noise |
| Gradient Boosting | High accuracy | Prone to overfitting |

---

### **q) Bias & Variance**  
- **Bias**: Model oversimplifies data (e.g., linear regression).  
- **Variance**: Model overfits (e.g., deep trees).  

---

### **r) Supervised Learning**  
Model learns from labeled data (e.g., Classification, Regression).  

---

### **s) Logistic Regression**  
Uses the sigmoid function to classify probabilities into two categories.  

---

### **t) Precision, Recall & F1 Score**  
\[
Precision = \frac{TP}{TP + FP}, \quad Recall = \frac{TP}{TP + FN}, \quad F1 = 2 \times \frac{Precision \times Recall}{Precision + Recall}
\]

---

### **u) How Random Forest Handles High Variance**  
Averaging predictions across multiple trees reduces variance without increasing bias.  

---

### **v) Steps in KNN Algorithm**  
1. Choose \( K \).  
2. Calculate distances.  
3. Select nearest neighbors.  
4. Assign majority class.  

---

### **w) Bagging vs. Boosting**  
- **Bagging**: Parallel training to reduce variance.  
- **Boosting**: Sequential training to correct mistakes.  

---

### **x) Binomial Logistic Regression vs. Linear Regression**  
- **Binomial Logistic Regression:** Used for binary classification.  
- **Linear Regression:** Predicts continuous outcomes.  

---

### **y) Coefficients in Logistic Regression**  
Indicate the effect of a feature on the log-odds of the outcome.  

---

### **z) Proximity Measures in KNN**  
- **Euclidean Distance:** Most common.  
- **Manhattan Distance:** Used when axes are non-uniform.  

---

### **aa) Purity of a Node in Decision Trees**  
Measured using **Gini Impurity** or **Entropy**.  

---

### **ab) AdaBoost vs. Gradient Boosting**  
| Feature | AdaBoost | Gradient Boosting |
|------------|---------|------------------|
| Training | Weighted samples | Residual learning |
| Complexity | Simple | Complex |
| Performance | Fast | More accurate |


    for col in cat.columns
    data[col].nunique()
    'count': data[c] .value_counts(),
    'percent': data[c] .value_counts(normalize=True)*100
    sns.countplot(data[col],label="count")
    plt.show()

    # Plot box plots for numerical variables to visualize outliers
    numerical_columns = data.select_dtypes(include=['int64', 'float64']).columns
    print(numerical_columns)
    i=0
    plt.figure(figsize=(15, 10))
    for column in numerical_columns:
    i=i+1
    plt.subplot(8, 4, i)
    sns.boxplot(data=data[column])
    plt.title(column)

    plt.tight_layout()
    plt.show()

    y = df_employee["Attrition"].map({"Yes": 1, "No": 0})  # Convert target to binary format

    df_employee['JobLevel']=df_employee['JobLevel'].astype('object')
    X = pd.get_dummies(df_employee.drop("Attrition", axis=1), drop_first=True)

    # Assuming 'df' is your DataFrame and 'y' is the target variable
    X = df_bank.drop('y', axis=1)  # Features
    y = df_bank['y']  # Target

    # Train a RandomForestClassifier
    model = RandomForestClassifier(random_state=42)
    model.fit(X, y)


    from sklearn.metrics import
    accuracy_score,precision_score.average_precision_score
    ,recall_score,f1_score

    accuracy = accuracy_score(y_test, y_pred)
    print(classification_report(y_test, y_pred))
    print(confusion_matrix(y_test, y_pred))

    feature_importances = pd.Series(model.feature_importances_, index=X.columns)
    feature_importances.sort_values(ascending=False, inplace=True)

    # Print and plot feature importances
    print("Feature Importances:\n")
    print(feature_importances)

    feature_importances.plot(kind='bar', title='Feature Importance', figsize=(10, 6))
    plt.show()



    # Get the coefficients and calculate odds ratios
    coefficients = model.coef_[0]
    odds_ratios = np.exp(coefficients)

    # Create a DataFrame to display features, coefficients, and odds ratios
    feature_summary = pd.DataFrame({

    'Coefficient (β)': coefficients,
    'Odds Ratio (e^β)': odds_ratios
    })

    # Confusion Matrix Calculation
    conf_matrix = confusion_matrix(y_test, y_pred)
    tn, fp, fn, tp = conf_matrix.ravel()

    # Calculate Sensitivity, Specificity, Misclassification Rate, and Accuracy
    sensitivity = tp / (tp + fn)
    specificity = tn / (tn + fp)
    misclassification_rate = (fp + fn) / len(y_test) * 100
    accuracy = accuracy_score(y_test, y_pred) * 100

    # Output the results
    print("Sensitivity (Recall):", sensitivity)
    print("Specificity:", specificity)
    print("Misclassification Rate (%):", misclassification_rate)
    print("Accuracy (%):", accuracy)


    cm = confusion_matrix(y_test, y_pred)
    sns.heatmap(cm, annot=True, fmt="d", cmap="Blues", xticklabels=labels, yticklabels=labels)
    disp = ConfusionMatrixDisplay(confusion_matrix=cm, display_labels=model.classes_)


    y_proba = model.predict_proba(X_test)[:, 1]  # Probabilities for the positive class
    roc_auc = roc_auc_score(y_test, y_proba)
    print(f"ROC AUC Score: {roc_auc:.2f}")

    # Plot the ROC curve
    fpr, tpr, thresholds = roc_curve(y_test, y_proba)
    plt.figure()
    plt.plot(fpr, tpr, color='blue', label=f'ROC Curve (AUC = {roc_auc:.2f})')
    plt.plot([0, 1], [0, 1], color='gray', linestyle='--')  # Diagonal line for random chance
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver Operating Characteristic Curve')
    plt.legend(loc='lower right')
    plt.show()


    print("Logistic Regression Coefficients and Odds Ratios:")
    print(feature_summary)

    from sklearn.metrics import log_loss

    # Calculate log-likelihood for the fitted model on the training set
    log_likelihood_fitted = -log_loss(y_train, model.predict_proba(X_train), normalize=False)

    # Calculate log-likelihood for the null model
    # The null model assumes only the intercept (mean probability of y) without any predictors
    y_null = np.full(y_train.shape, y_train.mean())  # Mean probability of y
    log_likelihood_null = -log_loss(y_train, y_null, normalize=False)

    # Calculate McFadden's R-squared
    mcfadden_r_squared = 1 - (log_likelihood_fitted / log_likelihood_null)

    print("McFadden's R-squared:", mcfadden_r_squared)


    cv_scores = cross_val_score(model, X, y, cv=10, scoring='accuracy')

    # Calculate the average accuracy
    average_accuracy = np.mean(cv_scores)

    # Print the results
    print(f"Cross-Validation Scores: {cv_scores}")
    print(f"Average Accuracy: {average_accuracy:.2f}")


    1. LogisticRegression(random_state=42))
    logreg = LogisticRegression(solver='lbfgs')
    2 model = GaussianNB()
    3 knn = KNeighborsClassifier(n_neighbors=5, metric='euclidean')
    4 model = DecisionTreeClassifier(random_state=42)
        decision_tree = DecisionTreeClassifier(random_state=42,max_features = i, max_leaf_nodes=5)
    5 model = RandomForestClassifier(n_estimators = i, random_state=42)

    1. **[ADABoost](#ada)**
    2. **[Gradient Boosting](#gb)**
    3. **[Extreme Gradient Boosting (XGB)](#xgb)**


    # Define the hyperparameter grid
    param_grid = {
    'criterion': ['gini', 'entropy'],              # Criterion for split quality
    'max_depth': [None, 10, 20, 30, 40, 50],       # Maximum depth of the tree
    'min_samples_split': [2, 5, 10, 20]            # Minimum number of samples required to split an internal node
    }

    # Initialize the Decision Tree model
    decision_tree = DecisionTreeClassifier(random_state=42)

    # Perform Grid Search with Cross-Validation
    grid_search = GridSearchCV(estimator=decision_tree, param_grid=param_grid, cv=10, scoring='accuracy')
    grid_search.fit(X, y)

    # Get the best parameters and best score
    best_params = grid_search.best_params_
    best_score = grid_search.best_score_

    # Output the results
    print("Best Hyperparameters:", best_params)
    print("Best Cross-Validated Accuracy:", best_score)

'''